@@ -39,7 +39,6 @@ Created 3/26/1996 Heikki Tuuri
 #include "row0row.h"
 #include "row0mysql.h"
 #include "row0ins.h"
-#include "mariadb_stats.h"
 
 /** The search tuple corresponding to TRX_UNDO_INSERT_METADATA. */
 const dtuple_t trx_undo_metadata = {
@@ -152,7 +151,9 @@ trx_undo_log_v_idx(
 	ulint n_idx = 0;
 	for (const auto& v_index : vcol->v_indexes) {
 		n_idx++;
-		/* FIXME: index->id is 64 bits! */
+		if (uint32_t hi= uint32_t(v_index.index->id >> 32)) {
+			size += 1 + mach_get_compressed_size(hi);
+		}
 		size += mach_get_compressed_size(uint32_t(v_index.index->id));
 		size += mach_get_compressed_size(v_index.nth_field);
 	}
@@ -179,10 +180,14 @@ trx_undo_log_v_idx(
 	ptr += mach_write_compressed(ptr, n_idx);
 
 	for (const auto& v_index : vcol->v_indexes) {
-		ptr += mach_write_compressed(
-			/* FIXME: index->id is 64 bits! */
-			ptr, uint32_t(v_index.index->id));
-
+		/* This is compatible with
+		ptr += mach_u64_write_much_compressed(ptr, v_index.index-id)
+		(the added "if" statement is fixing an old regression). */
+		if (uint32_t hi= uint32_t(v_index.index->id >> 32)) {
+			*ptr++ = 0xff;
+			ptr += mach_write_compressed(ptr, hi);
+		}
+		ptr += mach_write_compressed(ptr, uint32_t(v_index.index->id));
 		ptr += mach_write_compressed(ptr, v_index.nth_field);
 	}
 
@@ -221,7 +226,15 @@ trx_undo_read_v_idx_low(
 	dict_index_t*	clust_index = dict_table_get_first_index(table);
 
 	for (ulint i = 0; i < num_idx; i++) {
-		index_id_t	id = mach_read_next_compressed(&ptr);
+		index_id_t	id = 0;
+		/* This is like mach_u64_read_much_compressed(),
+		but advancing ptr to the next field. */
+		if (*ptr == 0xff) {
+			ptr++;
+			id = mach_read_next_compressed(&ptr);
+			id <<= 32;
+		}
+		id |= mach_read_next_compressed(&ptr);
 		ulint		pos = mach_read_next_compressed(&ptr);
 		dict_index_t*	index = dict_table_get_next_index(clust_index);
 
@@ -388,7 +401,6 @@ static
 uint16_t
 trx_undo_page_report_insert(
 	buf_block_t*	undo_block,
-	trx_t*		trx,
 	dict_index_t*	index,
 	const dtuple_t*	clust_entry,
 	mtr_t*		mtr,
@@ -418,7 +430,7 @@ trx_undo_page_report_insert(
 
 	/* Store first some general parameters to the undo log */
 	*ptr++ = TRX_UNDO_INSERT_REC;
-	ptr += mach_u64_write_much_compressed(ptr, trx->undo_no);
+	ptr += mach_u64_write_much_compressed(ptr, mtr->trx->undo_no);
 	ptr += mach_u64_write_much_compressed(ptr, index->table->id);
 
 	if (write_empty) {
@@ -772,7 +784,6 @@ uint16_t
 trx_undo_page_report_modify(
 /*========================*/
 	buf_block_t*	undo_block,	/*!< in: undo log page */
-	trx_t*		trx,		/*!< in: transaction */
 	dict_index_t*	index,		/*!< in: clustered index where update or
 					delete marking is done */
 	const rec_t*	rec,		/*!< in: clustered index record which
@@ -820,12 +831,16 @@ trx_undo_page_report_modify(
 	byte*		type_cmpl_ptr;
 	ulint		i;
 	trx_id_t	trx_id;
-	ibool		ignore_prefix = FALSE;
+	bool		ignore_prefix = false;
 	byte		ext_buf[REC_VERSION_56_MAX_INDEX_COL_LEN
 				+ BTR_EXTERN_FIELD_REF_SIZE];
 	bool		first_v_col = true;
 
 	/* Store first some general parameters to the undo log */
+	field = rec_get_nth_field(rec, offsets, index->db_trx_id(), &flen);
+	ut_ad(flen == DATA_TRX_ID_LEN);
+
+	trx_id = trx_read_trx_id(field);
 
 	if (!update) {
 		ut_ad(!rec_is_delete_marked(rec, dict_table_is_comp(table)));
@@ -833,14 +848,15 @@ trx_undo_page_report_modify(
 	} else if (rec_is_delete_marked(rec, dict_table_is_comp(table))) {
 		/* In delete-marked records, DB_TRX_ID must
 		always refer to an existing update_undo log record. */
-		ut_ad(row_get_rec_trx_id(rec, index, offsets));
+		ut_ad(trx_id);
 
 		type_cmpl = TRX_UNDO_UPD_DEL_REC;
+
 		/* We are about to update a delete marked record.
-		We don't typically need the prefix in this case unless
+		We don't typically need a BLOB prefix in this case unless
 		the delete marking is done by the same transaction
 		(which we check below). */
-		ignore_prefix = TRUE;
+		ignore_prefix = trx_id != mtr->trx->id;
 	} else {
 		type_cmpl = TRX_UNDO_UPD_EXIST_REC;
 	}
@@ -849,7 +865,7 @@ trx_undo_page_report_modify(
 	type_cmpl_ptr = ptr;
 
 	*ptr++ = (byte) type_cmpl;
-	ptr += mach_u64_write_much_compressed(ptr, trx->undo_no);
+	ptr += mach_u64_write_much_compressed(ptr, mtr->trx->undo_no);
 
 	ptr += mach_u64_write_much_compressed(ptr, table->id);
 
@@ -859,18 +875,6 @@ trx_undo_page_report_modify(
 	*ptr++ = (byte) rec_get_info_bits(rec, dict_table_is_comp(table));
 
 	/* Store the values of the system columns */
-	field = rec_get_nth_field(rec, offsets, index->db_trx_id(), &flen);
-	ut_ad(flen == DATA_TRX_ID_LEN);
-
-	trx_id = trx_read_trx_id(field);
-
-	/* If it is an update of a delete marked record, then we are
-	allowed to ignore blob prefixes if the delete marking was done
-	by some other trx as it must have committed by now for us to
-	allow an over-write. */
-	if (trx_id == trx->id) {
-		ignore_prefix = false;
-	}
 	ptr += mach_u64_write_compressed(ptr, trx_id);
 
 	field = rec_get_nth_field(rec, offsets, index->db_roll_ptr(), &flen);
@@ -1208,9 +1212,12 @@ store_len:
 				columns that were updated. */
 
 				for (i = 0; i < update->n_fields; i++) {
+					const upd_field_t* fld =
+						upd_get_nth_field(update, i);
+					if (upd_fld_is_virtual_col(fld))
+						continue;
 					const ulint field_no
-						= upd_get_nth_field(update, i)
-						->field_no;
+						= fld->field_no;
 					if (field_no >= index->n_fields
 					    || dict_index_get_nth_field(
 						    index, field_no)->col
@@ -1510,7 +1517,7 @@ trx_undo_update_rec_get_update(
 				&field_no);
 			first_v_col = false;
 			/* This column could be dropped or no longer indexed */
-			if (field_no >= index->n_fields) {
+			if (field_no == FIL_NULL) {
 				/* Mark this is no longer needed */
 				upd_field->field_no = REC_MAX_N_FIELDS;
 
@@ -1663,7 +1670,6 @@ trx_undo_update_rec_get_update(
 }
 
 /** Report a RENAME TABLE operation.
-@param[in,out]	trx	transaction
 @param[in]	table	table that is being renamed
 @param[in,out]	block	undo page
 @param[in,out]	mtr	mini-transaction
@@ -1671,7 +1677,7 @@ trx_undo_update_rec_get_update(
 @retval	0	in case of failure */
 static
 uint16_t
-trx_undo_page_report_rename(trx_t* trx, const dict_table_t* table,
+trx_undo_page_report_rename(const dict_table_t* table,
 			    buf_block_t* block, mtr_t* mtr)
 {
 	byte*	ptr_first_free  = my_assume_aligned<2>(TRX_UNDO_PAGE_HDR
@@ -1697,7 +1703,7 @@ trx_undo_page_report_rename(trx_t* trx, const dict_table_t* table,
 
 	byte* ptr = start + 2;
 	*ptr++ = TRX_UNDO_RENAME_TABLE;
-	ptr += mach_u64_write_much_compressed(ptr, trx->undo_no);
+	ptr += mach_u64_write_much_compressed(ptr, mtr->trx->undo_no);
 	ptr += mach_u64_write_much_compressed(ptr, table->id);
 	memcpy(ptr, table->name.m_name, len);
 	ptr += len;
@@ -1718,10 +1724,10 @@ dberr_t trx_undo_report_rename(trx_t* trx, const dict_table_t* table)
 	ut_ad(trx->id);
 	ut_ad(!table->is_temporary());
 
-	mtr_t		mtr;
+	mtr_t		mtr{trx};
 	dberr_t		err;
 	mtr.start();
-	if (buf_block_t* block = trx_undo_assign(trx, &err, &mtr)) {
+	if (buf_block_t* block = trx_undo_assign(&mtr, &err)) {
 		trx_undo_t*	undo = trx->rsegs.m_redo.undo;
 		ut_ad(err == DB_SUCCESS);
 		ut_ad(undo);
@@ -1731,7 +1737,7 @@ dberr_t trx_undo_report_rename(trx_t* trx, const dict_table_t* table)
 			      == block->page.id().page_no());
 
 			if (uint16_t offset = trx_undo_page_report_rename(
-				    trx, table, block, &mtr)) {
+				    table, block, &mtr)) {
 				undo->top_page_no = undo->last_page_no;
 				undo->top_offset  = offset;
 				undo->top_undo_no = trx->undo_no++;
@@ -1863,9 +1869,11 @@ trx_undo_report_row_operation(
 	} else if (!m.second || !trx->bulk_insert) {
 		bulk = false;
 	} else if (index->table->is_temporary()) {
-	} else if (trx_has_lock_x(*trx, *index->table)
-		   && index->table->bulk_trx_id == trx->id) {
-		m.first->second.start_bulk_insert(index->table);
+	} else if (index->table->bulk_trx_id == trx->id
+		   && trx_has_lock_x(*trx, *index->table)) {
+		m.first->second.start_bulk_insert(
+			index->table,
+			thd_sql_command(trx->mysql_thd) != SQLCOM_LOAD);
 
 		if (dberr_t err = m.first->second.bulk_insert_buffered(
 			    *clust_entry, *index, trx)) {
@@ -1875,7 +1883,7 @@ trx_undo_report_row_operation(
 		bulk = false;
 	}
 
-	mtr_t		mtr;
+	mtr_t		mtr{trx};
 	dberr_t		err;
 	mtr.start();
 	trx_undo_t**	pundo;
@@ -1887,15 +1895,15 @@ trx_undo_report_row_operation(
 		mtr.set_log_mode(MTR_LOG_NO_REDO);
 		rseg = trx->get_temp_rseg();
 		pundo = &trx->rsegs.m_noredo.undo;
-		undo_block = trx_undo_assign_low<true>(trx, rseg, pundo,
-						       &mtr, &err);
+		undo_block = trx_undo_assign_low<true>(&mtr, &err,
+						       rseg, pundo);
 	} else {
 		ut_ad(!trx->read_only);
 		ut_ad(trx->id);
 		pundo = &trx->rsegs.m_redo.undo;
 		rseg = trx->rsegs.m_redo.rseg;
-		undo_block = trx_undo_assign_low<false>(trx, rseg, pundo,
-							&mtr, &err);
+		undo_block = trx_undo_assign_low<false>(&mtr, &err,
+							rseg, pundo);
 	}
 
 	trx_undo_t*	undo	= *pundo;
@@ -1911,10 +1919,10 @@ err_exit:
 	do {
 		uint16_t offset = !rec
 			? trx_undo_page_report_insert(
-				undo_block, trx, index, clust_entry, &mtr,
+				undo_block, index, clust_entry, &mtr,
 				bulk)
 			: trx_undo_page_report_modify(
-				undo_block, trx, index, rec, offsets, update,
+				undo_block, index, rec, offsets, update,
 				cmpl_info, clust_entry, &mtr);
 
 		if (UNIV_UNLIKELY(offset == 0)) {
@@ -2047,12 +2055,14 @@ err_exit:
 
 static dberr_t trx_undo_prev_version(const rec_t *rec, dict_index_t *index,
                                      rec_offs *offsets, mem_heap_t *heap,
-                                     rec_t **old_vers, mem_heap_t *v_heap,
-                                     dtuple_t **vrow, ulint v_status,
+                                     rec_t **old_vers,
+                                     const purge_sys_t::view_guard &check,
+                                     ulint v_status,
+                                     mem_heap_t *v_heap, dtuple_t **vrow,
                                      const trx_undo_rec_t *undo_rec);
 
 inline const buf_block_t *
-purge_sys_t::view_guard::get(const page_id_t id, mtr_t *mtr)
+purge_sys_t::view_guard::get(const page_id_t id, trx_t *trx, mtr_t *mtr)
 {
   buf_block_t *block;
   ut_ad(mtr->is_active());
@@ -2066,7 +2076,7 @@ purge_sys_t::view_guard::get(const page_id_t id, mtr_t *mtr)
       return block;
     }
   }
-  block= buf_pool.page_fix(id);
+  block= buf_pool.page_fix(id, trx);
   if (block)
   {
     mtr->memo_push(block, MTR_MEMO_BUF_FIX);
@@ -2099,7 +2109,6 @@ must hold a latch on the index page of the clustered index record.
 @retval DB_SUCCESS if previous version was successfully built,
 or if it was an insert or the undo record refers to the table before rebuild
 @retval DB_MISSING_HISTORY if the history is missing */
-TRANSACTIONAL_TARGET
 dberr_t trx_undo_prev_version_build(const rec_t *rec, dict_index_t *index,
                                     rec_offs *offsets, mem_heap_t *heap,
                                     rec_t **old_vers, mtr_t *mtr,
@@ -2124,7 +2133,9 @@ dberr_t trx_undo_prev_version_build(const rec_t *rec, dict_index_t *index,
 
   ut_ad(!index->table->skip_alter_undo);
 
-  mariadb_increment_undo_records_read();
+  if (!mtr->trx);
+  else if (ha_handler_stats *stats= mtr->trx->active_handler_stats)
+    stats->undo_records_read++;
   const auto savepoint= mtr->get_savepoint();
   dberr_t err= DB_MISSING_HISTORY;
   purge_sys_t::view_guard check{v_status == TRX_UNDO_CHECK_PURGE_PAGES
@@ -2140,7 +2151,7 @@ dberr_t trx_undo_prev_version_build(const rec_t *rec, dict_index_t *index,
     if (const buf_block_t *undo_page=
         check.get(page_id_t{trx_sys.rseg_array[(roll_ptr >> 48) & 0x7f].
                             space->id,
-                            uint32_t(roll_ptr >> 16)}, mtr))
+                            uint32_t(roll_ptr >> 16)}, mtr->trx, mtr))
     {
       static_assert(ROLL_PTR_BYTE_POS == 0, "");
       const uint16_t offset{uint16_t(roll_ptr)};
@@ -2149,7 +2160,8 @@ dberr_t trx_undo_prev_version_build(const rec_t *rec, dict_index_t *index,
       if (UNIV_UNLIKELY(end > offset &&
                         end < srv_page_size - FIL_PAGE_DATA_END))
         err= trx_undo_prev_version(rec, index, offsets, heap,
-                                   old_vers, v_heap, vrow, v_status, undo_rec);
+                                   old_vers, check, v_status, v_heap, vrow,
+                                   undo_rec);
     }
   }
 
@@ -2159,8 +2171,10 @@ dberr_t trx_undo_prev_version_build(const rec_t *rec, dict_index_t *index,
 
 static dberr_t trx_undo_prev_version(const rec_t *rec, dict_index_t *index,
                                      rec_offs *offsets, mem_heap_t *heap,
-                                     rec_t **old_vers, mem_heap_t *v_heap,
-                                     dtuple_t **vrow, ulint v_status,
+                                     rec_t **old_vers,
+                                     const purge_sys_t::view_guard &check,
+                                     ulint v_status,
+                                     mem_heap_t *v_heap, dtuple_t **vrow,
                                      const trx_undo_rec_t *undo_rec)
 {
 	byte type, cmpl_info;
@@ -2217,6 +2231,21 @@ static dberr_t trx_undo_prev_version(const rec_t *rec, dict_index_t *index,
 	byte* buf;
 
 	if (row_upd_changes_field_size_or_external(index, offsets, update)) {
+		/* When CHECK TABLE ... EXTENDED checks for orphan
+		records in secondary indexes, it normally covers some
+		history that is already being purged. This is safe as
+		long as the undo log records have not been freed yet.
+
+		However, BLOBs are only safe to access as long as the
+		purge_sys.view does not permit them to be freed. The
+		check.latch will freeze the purge_sys.view by blocking
+		purge_sys.clone_oldest_view() at the start of
+		trx_purge() or by blocking purge_sys.batch_cleanup()
+		at the end of trx_purge(). */
+		if (check.is_extended() && purge_sys.is_purgeable(trx_id)) {
+			return DB_SUCCESS;
+		}
+
 		/* We should confirm the existence of disowned external data,
 		if the previous version record is delete marked. If the trx_id
 		of the previous record is seen by purge view, we should treat
@@ -2230,7 +2259,7 @@ static dberr_t trx_undo_prev_version(const rec_t *rec, dict_index_t *index,
 		the BLOB. */
 
 		if (update->info_bits & REC_INFO_DELETED_FLAG
-		    && purge_sys.is_purgeable(trx_id)) {
+		    && check.view().changes_visible(trx_id)) {
 			return DB_SUCCESS;
 		}
 
